#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections import defaultdict
from datetime import datetime

import attr
import dateutil
import pandas as pd
import pytz
import trafaret as t

from datarobot._compat import String
from datarobot.enums import DEFAULT_MAX_WAIT
from datarobot.models.api_object import APIObject
from datarobot.models.custom_inference_image import CustomInferenceImage
from datarobot.models.custom_model_version import CustomModelVersion

from ..helpers.deployment_monitoring import DeploymentQueryBuilderMixin
from ..utils import deprecated, encode_utf8_if_py2, from_api, get_id_from_location
from ..utils.pagination import unpaginate
from ..utils.waiters import wait_for_async_resolution
from .accuracy import Accuracy, AccuracyOverTime
from .data_drift import FeatureDrift, TargetDrift
from .service_stats import ServiceStats, ServiceStatsOverTime


class Deployment(APIObject, DeploymentQueryBuilderMixin):
    """A deployment created from a DataRobot model.

    Attributes
    ----------
    id : str
        the id of the deployment
    label : str
        the label of the deployment
    description : str
        the description of the deployment
    default_prediction_server : dict
        information on the default prediction server of the deployment
    model : dict
        information on the model of the deployment
    capabilities : dict
        information on the capabilities of the deployment
    prediction_usage : dict
        information on the prediction usage of the deployment
    permissions : list
        (New in version v2.18) user's permissions on the deployment
    service_health : dict
        information on the service health of the deployment
    model_health : dict
        information on the model health of the deployment
    accuracy_health : dict
        information on the accuracy health of the deployment
    """

    _path = "deployments/"
    _default_prediction_server_converter = t.Dict(
        {
            t.Key("id", optional=True): String(allow_blank=True),
            t.Key("url", optional=True): String(allow_blank=True),
            t.Key("datarobot-key", optional=True): String(allow_blank=True),
        }
    ).allow_extra("*")
    _model_converter = t.Dict(
        {
            t.Key("id", optional=True): String(),
            t.Key("type", optional=True): String(allow_blank=True),
            t.Key("target_name", optional=True): String(allow_blank=True),
            t.Key("project_id", optional=True): String(allow_blank=True),
        }
    ).allow_extra("*")
    _capabilities = t.Dict(
        {
            t.Key("supports_drift_tracking", optional=True): t.Bool(),
            t.Key("supports_model_replacement", optional=True): t.Bool(),
        }
    ).allow_extra("*")
    _prediction_usage = t.Dict(
        {
            t.Key("daily_rates", optional=True): t.List(t.Float()),
            t.Key("last_timestamp", optional=True): String >> (lambda s: dateutil.parser.parse(s)),
        }
    ).allow_extra("*")
    _health = t.Dict(
        {
            t.Key("status", optional=True): String(allow_blank=True),
            t.Key("message", optional=True): String(allow_blank=True),
            t.Key("start_date", optional=True): String >> (lambda s: dateutil.parser.parse(s)),
            t.Key("end_date", optional=True): String >> (lambda s: dateutil.parser.parse(s)),
        }
    ).allow_extra("*")
    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("label", optional=True): String(allow_blank=True),
            t.Key("description", optional=True): String(allow_blank=True) | t.Null(),
            t.Key("default_prediction_server", optional=True): _default_prediction_server_converter,
            t.Key("model", optional=True): _model_converter,
            t.Key("capabilities", optional=True): _capabilities,
            t.Key("prediction_usage", optional=True): _prediction_usage,
            t.Key("permissions", optional=True): t.List(String),
            t.Key("service_health", optional=True): _health,
            t.Key("model_health", optional=True): _health,
            t.Key("accuracy_health", optional=True): _health,
        }
    ).allow_extra("*")

    def __init__(
        self,
        id=None,
        label=None,
        description=None,
        default_prediction_server=None,
        model=None,
        capabilities=None,
        prediction_usage=None,
        permissions=None,
        service_health=None,
        model_health=None,
        accuracy_health=None,
    ):
        self.id = id
        self.label = label
        self.description = description
        self.default_prediction_server = default_prediction_server
        self.model = model
        self.capabilities = capabilities
        self.prediction_usage = prediction_usage
        self.permissions = permissions
        self.service_health = service_health
        self.model_health = model_health
        self.accuracy_health = accuracy_health

    def __repr__(self):
        return encode_utf8_if_py2(u"{}({})".format(self.__class__.__name__, self.label or self.id))

    @classmethod
    def create_from_learning_model(
        cls, model_id, label, description=None, default_prediction_server_id=None
    ):
        """Create a deployment from a DataRobot model.

        .. versionadded:: v2.17

        Parameters
        ----------
        model_id : str
            id of the DataRobot model to deploy
        label : str
            a human readable label of the deployment
        description : str, optional
            a human readable description of the deployment
        default_prediction_server_id : str, optional
            an identifier of a prediction server to be used as the default prediction server

        Returns
        -------
        deployment : Deployment
            The created deployment

        Examples
        --------
        .. code-block:: python

            from datarobot import Project, Deployment
            project = Project.get('5506fcd38bd88f5953219da0')
            model = project.get_models()[0]
            deployment = Deployment.create_from_learning_model(model.id, 'New Deployment')
            deployment
            >>> Deployment('New Deployment')
        """

        payload = {"model_id": model_id, "label": label, "description": description}
        if default_prediction_server_id:
            payload["default_prediction_server_id"] = default_prediction_server_id

        url = "{}fromLearningModel/".format(cls._path)
        deployment_id = cls._client.post(url, data=payload).json()["id"]
        return cls.get(deployment_id)

    @classmethod
    def _create_from_custom_model_entity(
        cls,
        custom_model_entity_id,
        label,
        entity_type,
        description=None,
        default_prediction_server_id=None,
        max_wait=DEFAULT_MAX_WAIT,
    ):
        # check if model package of the custom model image is already created
        existing_model_packages = unpaginate(
            "modelPackages/", {"model_id": custom_model_entity_id}, cls._client
        )

        try:
            model_package_id = next(existing_model_packages)["id"]
        except StopIteration:
            # model package of the custom model entity does not exists,
            # so create one
            if entity_type == CustomModelVersion.__name__:
                field_name = "custom_model_version_id"
                route = "fromCustomModelVersion"
            else:
                field_name = "custom_model_image_id"
                route = "fromCustomModelImage"
            model_package_payload = {field_name: custom_model_entity_id}

            model_package_id = cls._client.post(
                "modelPackages/{}/".format(route), data=model_package_payload
            ).json()["id"]

        # create deployment from the model package
        deployment_payload = {
            "model_package_id": model_package_id,
            "label": label,
            "description": description,
        }
        if default_prediction_server_id:
            deployment_payload["default_prediction_server_id"] = default_prediction_server_id
        response = cls._client.post(
            "{}fromModelPackage/".format(cls._path), data=deployment_payload
        )

        # wait for LRS job resolution to support making predictions against the deployment
        wait_for_async_resolution(cls._client, response.headers["Location"], max_wait)

        deployment_id = response.json()["id"]
        return cls.get(deployment_id)

    @classmethod
    @deprecated(
        deprecated_since_version="v2.23",
        will_remove_version="v2.24",
        message="The use of CustomModelImages is deprecated. "
        "Use create_from_custom_model_version instead.",
    )
    def create_from_custom_model_image(
        cls,
        custom_model_image_id,
        label,
        description=None,
        default_prediction_server_id=None,
        max_wait=DEFAULT_MAX_WAIT,
    ):
        """Create a deployment from a DataRobot custom model image.

        Parameters
        ----------
        custom_model_image_id : str
            id of the DataRobot custom model image to deploy
        label : str
            a human readable label of the deployment
        description : str, optional
            a human readable description of the deployment
        default_prediction_server_id : str, optional
            an identifier of a prediction server to be used as the default prediction server
        max_wait : int, optional
            seconds to wait for successful resolution of a deployment creation job.
            Deployment supports making predictions only after a deployment creating job
            has successfully finished

        Returns
        -------
        deployment : Deployment
            The created deployment
        """

        return cls._create_from_custom_model_entity(
            custom_model_entity_id=custom_model_image_id,
            label=label,
            entity_type=CustomInferenceImage.__name__,
            description=description,
            default_prediction_server_id=default_prediction_server_id,
            max_wait=max_wait,
        )

    @classmethod
    def create_from_custom_model_version(
        cls,
        custom_model_version_id,
        label,
        description=None,
        default_prediction_server_id=None,
        max_wait=DEFAULT_MAX_WAIT,
    ):
        """Create a deployment from a DataRobot custom model image.

        Parameters
        ----------
        custom_model_version_id : str
            id of the DataRobot custom model version to deploy
            The version must have a base_environment_id.
        label : str
            a human readable label of the deployment
        description : str, optional
            a human readable description of the deployment
        default_prediction_server_id : str, optional
            an identifier of a prediction server to be used as the default prediction server
        max_wait : int, optional
            seconds to wait for successful resolution of a deployment creation job.
            Deployment supports making predictions only after a deployment creating job
            has successfully finished

        Returns
        -------
        deployment : Deployment
            The created deployment
        """

        return cls._create_from_custom_model_entity(
            custom_model_entity_id=custom_model_version_id,
            label=label,
            entity_type=CustomModelVersion.__name__,
            description=description,
            default_prediction_server_id=default_prediction_server_id,
            max_wait=max_wait,
        )

    @classmethod
    def list(cls, order_by=None, search=None, filters=None):
        """List all deployments a user can view.

        .. versionadded:: v2.17

        Parameters
        ----------
        order_by : str, optional
            (New in version v2.18) the order to sort the deployment list by, defaults to `label`

            Allowed attributes to sort by are:

            * ``label``
            * ``serviceHealth``
            * ``modelHealth``
            * ``accuracyHealth``
            * ``recentPredictions``
            * ``lastPredictionTimestamp``

            If the sort attribute is preceded by a hyphen, deployments will be sorted in descending
            order, otherwise in ascending order.

            For health related sorting, ascending means failing, warning, passing, unknown.
        search : str, optional
            (New in version v2.18) case insensitive search against deployment's
            label and description.
        filters : datarobot.models.deployment.DeploymentListFilters, optional
            (New in version v2.20) an object containing all filters that you'd like to apply to the
            resulting list of deployments. See
            :class:`~datarobot.models.deployment.DeploymentListFilters` for details on usage.

        Returns
        -------
        deployments : list
            a list of deployments the user can view

        Examples
        --------
        .. code-block:: python

            from datarobot import Deployment
            deployments = Deployment.list()
            deployments
            >>> [Deployment('New Deployment'), Deployment('Previous Deployment')]

        .. code-block:: python

            from datarobot import Deployment
            from datarobot.enums import DEPLOYMENT_SERVICE_HEALTH
            filters = DeploymentListFilters(
                role='OWNER',
                service_health=[DEPLOYMENT_SERVICE_HEALTH.FAILING]
            )
            filtered_deployments = Deployment.list(filters=filters)
            filtered_deployments
            >>> [Deployment('Deployment I Own w/ Failing Service Health')]
        """
        if filters is None:
            filters = DeploymentListFilters()

        param = {}
        if order_by:
            param["order_by"] = order_by
        if search:
            param["search"] = search
        param.update(filters.construct_query_args())
        data = unpaginate(cls._path, param, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def get(cls, deployment_id):
        """Get information about a deployment.

        .. versionadded:: v2.17

        Parameters
        ----------
        deployment_id : str
            the id of the deployment

        Returns
        -------
        deployment : Deployment
            the queried deployment

        Examples
        --------
        .. code-block:: python

            from datarobot import Deployment
            deployment = Deployment.get(deployment_id='5c939e08962d741e34f609f0')
            deployment.id
            >>>'5c939e08962d741e34f609f0'
            deployment.label
            >>>'New Deployment'
        """

        path = "{}{}/".format(cls._path, deployment_id)
        return cls.from_location(path)

    def update(self, label=None, description=None):
        """
        Update the label and description of this deployment.

        .. versionadded:: v2.19
        """

        payload = {}
        if label:
            payload["label"] = label
        if description:
            payload["description"] = description
        if not payload:
            raise ValueError("")

        url = "{}{}/".format(self._path, self.id)
        self._client.patch(url, data=payload)

        if label:
            self.label = label
        if description:
            self.description = description

    def delete(self):
        """
        Delete this deployment.

        .. versionadded:: v2.17
        """

        url = "{}{}/".format(self._path, self.id)
        self._client.delete(url)

    def replace_model(self, new_model_id, reason, max_wait=600):
        """Replace the model used in this deployment. To confirm model replacement eligibility, use
         :meth:`~datarobot.Deployment.validate_replacement_model` beforehand.

        .. versionadded:: v2.17

        Model replacement is an asynchronous process, which means some preparatory work may
        be performed after the initial request is completed. This function will not return until all
        preparatory work is fully finished.

        Predictions made against this deployment will start using the new model as soon as the
        initial request is completed. There will be no interruption for predictions throughout
        the process.

        Parameters
        ----------
        new_model_id : str
            The id of the new model to use. If replacing the deployment's model with a
            CustomInferenceModel, a specific CustomModelVersion ID must be used.
        reason : MODEL_REPLACEMENT_REASON
            The reason for the model replacement. Must be one of 'ACCURACY', 'DATA_DRIFT', 'ERRORS',
            'SCHEDULED_REFRESH', 'SCORING_SPEED', or 'OTHER'. This value will be stored in the model
            history to keep track of why a model was replaced
        max_wait : int, optional
            (new in version 2.22) The maximum time to wait for
            model replacement job to complete before erroring

        Examples
        --------
        .. code-block:: python

            from datarobot import Deployment
            deployment = Deployment.get(deployment_id='5c939e08962d741e34f609f0')
            deployment.model['id'], deployment.model['type']
            >>>('5c0a979859b00004ba52e431', 'Decision Tree Classifier (Gini)')

            deployment.replace_model('5c0a969859b00004ba52e41b', MODEL_REPLACEMENT_REASON.ACCURACY)
            deployment.model['id'], deployment.model['type']
            >>>('5c0a969859b00004ba52e41b', 'Support Vector Classifier (Linear Kernel)')
        """

        url = "{}{}/model/".format(self._path, self.id)
        payload = {"modelId": new_model_id, "reason": reason}
        response = self._client.patch(url, data=payload)
        deployment_loc = wait_for_async_resolution(
            self._client, response.headers["Location"], max_wait=max_wait
        )
        deployment_id = get_id_from_location(deployment_loc)
        deployment = Deployment.get(deployment_id)
        self.model = deployment.model

        # Update prediction intervals settings and check if the new model can support them
        old_pred_int_settings = self.get_prediction_intervals_settings()
        try:
            if old_pred_int_settings["percentiles"]:
                self.update_prediction_intervals_settings(
                    percentiles=old_pred_int_settings["percentiles"],
                    enabled=old_pred_int_settings["enabled"],
                )
        except Exception:
            # Doing a catch-all here because any errors from prediction intervals should not affect
            # the rest of model replacement. If there are errors, then update deployment to use
            # default prediction intervals settings.
            self.update_prediction_intervals_settings(percentiles=[], enabled=False)

    def validate_replacement_model(self, new_model_id):
        """Validate a model can be used as the replacement model of the deployment.

        .. versionadded:: v2.17

        Parameters
        ----------
        new_model_id : str
            the id of the new model to validate

        Returns
        -------
        status : str
            status of the validation, will be one of 'passing', 'warning' or 'failing'.
            If the status is passing or warning, use :meth:`~datarobot.Deployment.replace_model` to
            perform a model replacement. If the status is failing, refer to ``checks`` for more
            detail on why the new model cannot be used as a replacement.
        message : str
            message for the validation result
        checks : dict
            explain why the new model can or cannot replace the deployment's current model
        """

        url = "{}{}/model/validation/".format(self._path, self.id)
        payload = {"modelId": new_model_id}
        data = from_api(self._client.post(url, data=payload).json())
        return data.get("status"), data.get("message"), data.get("checks")

    def get_features(self):
        """Retrieve the list of features needed to make predictions on this deployment.

        Notes
        -----

        Each `feature` dict contains the following structure:

        - ``name`` : str, feature name
        - ``feature_type`` : str, feature type
        - ``importance`` : float, numeric measure of the relationship strength between
          the feature and target (independent of model or other features)
        - ``date_format`` : str or None, the date format string for how this feature was
          interpreted, null if not a date feature, compatible with
          https://docs.python.org/2/library/time.html#time.strftime.
        - ``known_in_advance`` : bool, whether the feature was selected as known in advance in
          a time series model, false for non-time series models.

        Returns
        -------
        features: list
            a list of `feature` dict

        Examples
        --------
        .. code-block:: python

            from datarobot import Deployment
            deployment = Deployment.get(deployment_id='5c939e08962d741e34f609f0')
            features = deployment.get_features()
            features[0]['feature_type']
            >>>'Categorical'
            features[0]['importance']
            >>>0.133
        """
        url = "{}{}/features/".format(self._path, self.id)
        data = unpaginate(url, {}, self._client)
        return from_api([feature for feature in data], keep_null_keys=True)

    def submit_actuals(self, data, batch_size=10000):
        """Submit actuals for processing.
        The actuals submitted will be used to calculate accuracy metrics.

        Parameters
        ----------
        data: list or pandas.DataFrame
        batch_size: the max number of actuals in each request

        If `data` is a list, each item should be a dict-like object with the following keys and
        values; if `data` is a pandas.DataFrame, it should contain the following columns:

        - association_id: str, a unique identifier used with a prediction,
            max length 128 characters
        - actual_value: str or int or float, the actual value of a prediction;
            should be numeric for deployments with regression models or
            string for deployments with classification model
        - was_acted_on: bool, optional, indicates if the prediction was acted on in a way that
            could have affected the actual outcome
        - timestamp: datetime or string in RFC3339 format, optional. If the datetime provided
            does not have a timezone, we assume it is UTC.

        Raises
        ------
        ValueError
            if input data is not a list of dict-like objects or a pandas.DataFrame
            if input data is empty

        Examples
        --------
        .. code-block:: python

            from datarobot import Deployment, AccuracyOverTime
            deployment = Deployment.get(deployment_id='5c939e08962d741e34f609f0')
            data = [{
                'association_id': '439917',
                'actual_value': 'True',
                'was_acted_on': True
            }]
            deployment.submit_actuals(data)
        """

        if not isinstance(data, (list, pd.DataFrame)):
            raise ValueError(
                "data should be either a list of dict-like objects or a pandas.DataFrame"
            )

        if not isinstance(batch_size, int) or not batch_size >= 1:
            raise ValueError(
                "batch_size should be an integer and should be greater than or equals to one"
            )

        if isinstance(data, pd.DataFrame):
            data = data.to_dict(orient="records")

        if not data:
            raise ValueError("data should not be empty")

        url = "{}{}/actuals/fromJSON/".format(self._path, self.id)
        for offset in range(0, len(data), batch_size):
            batch = data[offset : offset + batch_size]
            payload = []
            for item in batch:
                actual = {
                    "associationId": item["association_id"],
                    "actualValue": item["actual_value"],
                }

                # format wasActedOn
                was_acted_on = item.get("was_acted_on")
                if not pd.isna(was_acted_on):
                    actual["wasActedOn"] = item["was_acted_on"]

                # format timestamp
                timestamp = item.get("timestamp")
                if timestamp and not pd.isna(timestamp):
                    timestamp = item["timestamp"]
                    if isinstance(timestamp, datetime):
                        if not timestamp.tzinfo:
                            timestamp = timestamp.replace(tzinfo=pytz.utc)
                        timestamp = timestamp.isoformat()
                    actual["timestamp"] = timestamp

                payload.append(actual)
            response = self._client.post(url, data={"data": payload})
            wait_for_async_resolution(self._client, response.headers["Location"])

    def get_predictions_by_forecast_date_settings(self):
        """Retrieve predictions by forecast date settings of this deployment.

        .. versionadded:: v2.27

        Returns
        -------
        settings : dict
            Predictions by forecast date settings of the deployment is a dict with the following
            format:

            enabled : bool
                Is ''True'' if predictions by forecast date is enabled for this deployment.
                To update this setting, see
                :meth:`~datarobot.Deployment.update_predictions_by_forecast_date_settings`

            column_name : string
                The column name in prediction datasets to be used as forecast date.

            datetime_format : string
                The datetime format of the forecast date column in prediction datasets.
        """

        url = "{}{}/settings/".format(self._path, self.id)
        response_json = from_api(self._client.get(url).json())
        return response_json.get("predictions_by_forecast_date")

    def update_predictions_by_forecast_date_settings(
        self,
        enable_predictions_by_forecast_date,
        forecast_date_column_name=None,
        forecast_date_format=None,
        max_wait=DEFAULT_MAX_WAIT,
    ):
        """Update predictions by forecast date settings of this deployment.

        .. versionadded:: v2.27

        Updating predictions by forecast date setting is an asynchronous process,
        which means some preparatory work may be performed after the initial request is completed.
        This function will not return until all preparatory work is fully finished.

        Example
        -------
        To set predictions by forecast date settings to the same default settings you see when using
        the DataRobot web application, you use your ''Deployment'' object like this:

            >>> deployment.update_predictions_by_forecast_date_settings(
            >>>    enable_predictions_by_forecast_date=True,
            >>>    forecast_date_column_name="date (actual)",
            >>>    forecast_date_format="%Y-%m-%d")

        Parameters
        ----------
        enable_predictions_by_forecast_date : bool
            set to ''True'' if predictions by forecast date is to be turned on or set to ''False''
            if predictions by forecast date is to be turned off.

        forecast_date_column_name: string, optional
            The column name in prediction datasets to be used as forecast date.
            If ''enable_predictions_by_forecast_date'' is set to ''False'',
            then the parameter will be ignored.

        forecast_date_format: string, optional
            The datetime format of the forecast date column in prediction datasets.
            If ''enable_predictions_by_forecast_date'' is set to ''False'',
            then the parameter will be ignored.

        max_wait : int, optional
            seconds to wait for successful
        """

        payload = defaultdict(dict)
        payload["predictions_by_forecast_date"]["enabled"] = enable_predictions_by_forecast_date
        if enable_predictions_by_forecast_date:
            assert forecast_date_column_name, (
                "Please specify 'forecast_date_column_name' or set "
                "'enable_predictions_by_forecast_date' to False"
            )
            assert forecast_date_format, (
                "Please specify 'forecast_date_format' or set "
                "'enable_predictions_by_forecast_date' to False"
            )
            payload["predictions_by_forecast_date"]["column_name"] = forecast_date_column_name
            payload["predictions_by_forecast_date"]["datetime_format"] = forecast_date_format

        url = "{}{}/settings/".format(self._path, self.id)
        response = self._client.patch(url, data=payload)
        wait_for_async_resolution(self._client, response.headers["Location"], max_wait)

    def get_challenger_models_settings(self):
        """Retrieve challenger models settings of this deployment.

        .. versionadded:: v2.27

        Returns
        -------
        settings : dict
            Challenger models settings of the deployment is a dict with the following format:

            enabled : bool
                Is ''True'' if challenger models is enabled for this deployment. To update
                existing ''challenger_models'' settings, see
                :meth:`~datarobot.Deployment.update_challenger_models_settings`
        """

        url = "{}{}/settings/".format(self._path, self.id)
        response_json = from_api(self._client.get(url).json())
        return response_json.get("challenger_models")

    def update_challenger_models_settings(
        self, challenger_models_enabled, max_wait=DEFAULT_MAX_WAIT
    ):
        """Update challenger models settings of this deployment.

        .. versionadded:: v2.27

        Updating challenger models setting is an asynchronous process, which means some preparatory
        work may be performed after the initial request is completed. This function will not return
        until all preparatory work is fully finished.

        Parameters
        ----------
        challenger_models_enabled : bool
            set to ''True'' if challenger models is to be turned on or set to ''False'' if
            challenger models is to be turned off
        max_wait : int, optional
            seconds to wait for successful resolution
        """

        payload = defaultdict(dict)
        payload["challenger_models"]["enabled"] = challenger_models_enabled
        url = "{}{}/settings/".format(self._path, self.id)
        response = self._client.patch(url, data=payload)
        wait_for_async_resolution(self._client, response.headers["Location"], max_wait)

    def get_segment_analysis_settings(self):
        """Retrieve segment analysis settings of this deployment.

        .. versionadded:: v2.27

        Returns
        -------
        settings : dict
            Segment analysis settings of the deployment containing two items with keys
            ``enabled`` and ``attributes``, which are further described below.

            enabled : bool
                Set to ''True'' if segment analysis is enabled for this deployment. To update
                existing setting, see
                :meth:`~datarobot.Deployment.update_segment_analysis_settings`

            attributes : list
                To create or update existing segment analysis attributes, see
                :meth:`~datarobot.Deployment.update_segment_analysis_settings`
        """

        url = "{}{}/settings/".format(self._path, self.id)
        response_json = from_api(self._client.get(url).json())
        return response_json.get("segment_analysis")

    def update_segment_analysis_settings(
        self, segment_analysis_enabled, segment_analysis_attributes=None, max_wait=DEFAULT_MAX_WAIT,
    ):
        """Update segment analysis settings of this deployment.

        .. versionadded:: v2.27

        Updating segment analysis setting is an asynchronous process, which means some preparatory
        work may be performed after the initial request is completed. This function will not return
        until all preparatory work is fully finished.

        Parameters
        ----------
        segment_analysis_enabled : bool
            set to ''True'' if segment analysis is to be turned on or set to ''False'' if
            segment analysis is to be turned off

        segment_analysis_attributes: list, optional
            A list of strings that gives the segment attributes selected for tracking.

        max_wait : int, optional
            seconds to wait for successful resolution
        """

        payload = defaultdict(dict)
        payload["segment_analysis"]["enabled"] = segment_analysis_enabled
        if segment_analysis_attributes:
            payload["segment_analysis"]["attributes"] = segment_analysis_attributes
        url = "{}{}/settings/".format(self._path, self.id)
        response = self._client.patch(url, data=payload)
        wait_for_async_resolution(self._client, response.headers["Location"], max_wait)

    def get_drift_tracking_settings(self):
        """Retrieve drift tracking settings of this deployment.

        .. versionadded:: v2.17

        Returns
        -------
        settings : dict
            Drift tracking settings of the deployment containing two nested dicts with key
            ``target_drift`` and ``feature_drift``, which are further described below.

            ``Target drift`` setting contains:

            enabled : bool
                If target drift tracking is enabled for this deployment. To create or update
                existing ''target_drift'' settings, see
                :meth:`~datarobot.Deployment.update_drift_tracking_settings`

            ``Feature drift`` setting contains:

            enabled : bool
                If feature drift tracking is enabled for this deployment. To create or update
                existing ''feature_drift'' settings, see
                :meth:`~datarobot.Deployment.update_drift_tracking_settings`
        """

        url = "{}{}/settings/".format(self._path, self.id)
        response_json = from_api(self._client.get(url).json())
        return {
            key: value
            for key, value in response_json.items()
            if key in ["target_drift", "feature_drift"]
        }

    def update_drift_tracking_settings(
        self, target_drift_enabled=None, feature_drift_enabled=None, max_wait=DEFAULT_MAX_WAIT
    ):
        """Update drift tracking settings of this deployment.

        .. versionadded:: v2.17

        Updating drift tracking setting is an asynchronous process, which means some preparatory
        work may be performed after the initial request is completed. This function will not return
        until all preparatory work is fully finished.

        Parameters
        ----------
        target_drift_enabled : bool, optional
            if target drift tracking is to be turned on
        feature_drift_enabled : bool, optional
            if feature drift tracking is to be turned on
        max_wait : int, optional
            seconds to wait for successful resolution
        """

        payload = defaultdict(dict)
        if target_drift_enabled is not None:
            payload["targetDrift"]["enabled"] = target_drift_enabled
        if feature_drift_enabled is not None:
            payload["featureDrift"]["enabled"] = feature_drift_enabled
        if not payload:
            raise ValueError()

        url = "{}{}/settings/".format(self._path, self.id)
        response = self._client.patch(url, data=payload)
        wait_for_async_resolution(self._client, response.headers["Location"], max_wait)

    def get_association_id_settings(self):
        """Retrieve association ID setting for this deployment.

        .. versionadded:: v2.19

        Returns
        -------
        association_id_settings : dict in the following format:
            column_names : list[string], optional
                name of the columns to be used as association ID,
            required_in_prediction_requests : bool, optional
                whether the association ID column is required in prediction requests
        """

        url = "{}{}/settings/".format(self._path, self.id)
        response_json = from_api(self._client.get(url).json(), keep_null_keys=True)
        return response_json.get("association_id")

    def update_association_id_settings(
        self, column_names=None, required_in_prediction_requests=None, max_wait=DEFAULT_MAX_WAIT
    ):
        """Update association ID setting for this deployment.

        .. versionadded:: v2.19

        Parameters
        ----------
        column_names : list[string], optional
            name of the columns to be used as association ID,
            currently only support a list of one string
        required_in_prediction_requests : bool, optional
            whether the association ID column is required in prediction requests
        max_wait : int, optional
            seconds to wait for successful resolution
        """

        payload = defaultdict(dict)

        if column_names:
            payload["associationId"]["columnNames"] = column_names
        if required_in_prediction_requests is not None:
            payload["associationId"][
                "requiredInPredictionRequests"
            ] = required_in_prediction_requests
        if not payload:
            raise ValueError()

        url = "{}{}/settings/".format(self._path, self.id)
        response = self._client.patch(url, data=payload)
        wait_for_async_resolution(self._client, response.headers["Location"], max_wait)

    def get_predictions_data_collection_settings(self):
        """Retrieve predictions data collection settings of this deployment.

        .. versionadded:: v2.21

        Returns
        -------
        predictions_data_collection_settings : dict in the following format:
            enabled : bool
                If predictions data collection is enabled for this deployment. To update
                existing ''predictions_data_collection'' settings, see
                :meth:`~datarobot.Deployment.update_predictions_data_collection_settings`
        """

        url = "{}{}/settings/".format(self._path, self.id)
        response_json = from_api(self._client.get(url).json(), keep_null_keys=True)
        return response_json.get("predictions_data_collection")

    def update_predictions_data_collection_settings(self, enabled, max_wait=DEFAULT_MAX_WAIT):
        """Update predictions data collection settings of this deployment.

        .. versionadded:: v2.21

        Updating predictions data collection setting is an asynchronous process, which means some
        preparatory work may be performed after the initial request is completed.
        This function will not return until all preparatory work is fully finished.

        Parameters
        ----------
        enabled: bool
            if predictions data collection is to be turned on
        max_wait : int, optional
            seconds to wait for successful resolution
        """
        payload = {"predictionsDataCollection": {"enabled": enabled}}

        url = "{}{}/settings/".format(self._path, self.id)
        response = self._client.patch(url, data=payload)
        wait_for_async_resolution(self._client, response.headers["Location"], max_wait)

    def get_prediction_warning_settings(self):
        """Retrieve prediction warning settings of this deployment.

        .. versionadded:: v2.19

        Returns
        -------
        settings : dict in the following format:
            enabled : bool
                If target prediction_warning is enabled for this deployment. To create or update
                existing ''prediction_warning'' settings, see
                :meth:`~datarobot.Deployment.update_prediction_warning_settings`

            custom_boundaries : dict or None
                If None default boundaries for a model are used. Otherwise has following keys:
                    upper : float
                        All predictions greater than provided value are considered anomalous
                    lower : float
                        All predictions less than provided value are considered anomalous
        """

        url = "{}{}/settings/".format(self._path, self.id)
        response_json = from_api(self._client.get(url).json())
        return response_json.get("prediction_warning")

    def update_prediction_warning_settings(
        self,
        prediction_warning_enabled,
        use_default_boundaries=None,
        lower_boundary=None,
        upper_boundary=None,
        max_wait=DEFAULT_MAX_WAIT,
    ):
        """Update prediction warning settings of this deployment.

        .. versionadded:: v2.19

        Parameters
        ----------
        prediction_warning_enabled : bool
            If prediction warnings should be turned on.
        use_default_boundaries : bool, optional
            If default boundaries of the model should be used for the deployment.
        upper_boundary : float, optional
            All predictions greater than provided value will be considered anomalous
        lower_boundary : float, optional
            All predictions less than provided value will be considered anomalous
        max_wait : int, optional
            seconds to wait for successful resolution
        """

        payload = defaultdict(dict)
        payload["prediction_warning"]["enabled"] = prediction_warning_enabled
        if use_default_boundaries is True:
            payload["prediction_warning"]["custom_boundaries"] = None
        elif use_default_boundaries is False:
            if upper_boundary is not None and lower_boundary is not None:
                payload["prediction_warning"]["custom_boundaries"] = {
                    "upper": upper_boundary,
                    "lower": lower_boundary,
                }

        url = "{}{}/settings/".format(self._path, self.id)
        response = self._client.patch(url, data=payload, keep_attrs={"custom_boundaries"})
        wait_for_async_resolution(self._client, response.headers["Location"], max_wait)

    def get_prediction_intervals_settings(self):
        """ Retrieve prediction intervals settings for this deployment.

        .. versionadded:: v2.19

        Notes
        -----
        Note that prediction intervals are only supported for time series deployments.

        Returns
        -------
        dict in the following format:
            enabled : bool
                Whether prediction intervals are enabled for this deployment
            percentiles : list[int]
                List of enabled prediction intervals sizes for this deployment. Currently we only
                support one percentile at a time.
        """
        url = "{}{}/settings/".format(self._path, self.id)
        response_json = from_api(self._client.get(url).json())
        return response_json.get("prediction_intervals")

    def update_prediction_intervals_settings(
        self, percentiles, enabled=True, max_wait=DEFAULT_MAX_WAIT
    ):
        """Update prediction intervals settings for this deployment.

        .. versionadded:: v2.19

        Notes
        -----
        Updating prediction intervals settings is an asynchronous process, which means some
        preparatory work may be performed before the settings request is completed. This function
        will not return until all work is fully finished.

        Note that prediction intervals are only supported for time series deployments.

        Parameters
        ----------
        percentiles : list[int]
            The prediction intervals percentiles to enable for this deployment. Currently we only
            support setting one percentile at a time.
        enabled : bool, optional (defaults to True)
            Whether to enable showing prediction intervals in the results of predictions requested
            using this deployment.
        max_wait : int, optional
            seconds to wait for successful resolution

        Raises
        ------
        AssertionError
            If ``percentiles`` is in an invalid format
        AsyncFailureError
            If any of the responses from the server are unexpected
        AsyncProcessUnsuccessfulError
            If the prediction intervals calculation job has failed or has been cancelled.
        AsyncTimeoutError
            If the prediction intervals calculation job did not resolve in time
        """
        if percentiles:
            # Ensure percentiles is list[int] with length 1
            assert isinstance(percentiles, list) and len(percentiles) == 1

            # Make sure that the requested percentile is calculated
            from .model import DatetimeModel

            model = DatetimeModel(id=self.model["id"], project_id=self.model["project_id"])
            job = model.calculate_prediction_intervals(percentiles[0])
            job.wait_for_completion(max_wait)

        # Now update deployment with new prediction intervals settings
        payload = {"predictionIntervals": {"enabled": enabled, "percentiles": percentiles or []}}
        url = "{}{}/settings/".format(self._path, self.id)
        response = self._client.patch(url, data=payload)
        wait_for_async_resolution(self._client, response.headers["Location"], max_wait)

    def get_service_stats(
        self,
        model_id=None,
        start_time=None,
        end_time=None,
        execution_time_quantile=None,
        response_time_quantile=None,
        slow_requests_threshold=None,
    ):
        """Retrieve value of service stat metrics over a certain time period.

        .. versionadded:: v2.18

        Parameters
        ----------
        model_id : str, optional
            the id of the model
        start_time : datetime, optional
            start of the time period
        end_time : datetime, optional
            end of the time period
        execution_time_quantile : float, optional
            quantile for `executionTime`, defaults to 0.5
        response_time_quantile : float, optional
            quantile for `responseTime`, defaults to 0.5
        slow_requests_threshold : float, optional
            threshold for `slowRequests`, defaults to 1000

        Returns
        -------
        service_stats : ServiceStats
            the queried service stats metrics information
        """

        kwargs = {
            "model_id": model_id,
            "start_time": start_time,
            "end_time": end_time,
            "execution_time_quantile": execution_time_quantile,
            "response_time_quantile": response_time_quantile,
            "slow_requests_threshold": slow_requests_threshold,
        }
        return ServiceStats.get(self.id, **kwargs)

    def get_service_stats_over_time(
        self,
        metric=None,
        model_id=None,
        start_time=None,
        end_time=None,
        bucket_size=None,
        quantile=None,
        threshold=None,
    ):
        """Retrieve information about how a service stat metric changes over a certain time period.

        .. versionadded:: v2.18

        Parameters
        ----------
        metric : SERVICE_STAT_METRIC, optional
            the service stat metric to retrieve
        model_id : str, optional
            the id of the model
        start_time : datetime, optional
            start of the time period
        end_time : datetime, optional
            end of the time period
        bucket_size : str, optional
            time duration of a bucket, in ISO 8601 time duration format
        quantile : float, optional
            quantile for 'executionTime' or 'responseTime', ignored when querying other metrics
        threshold : int, optional
            threshold for 'slowQueries', ignored when querying other metrics

        Returns
        -------
        service_stats_over_time : ServiceStatsOverTime
            the queried service stats metric over time information
        """

        kwargs = {
            "metric": metric,
            "model_id": model_id,
            "start_time": start_time,
            "end_time": end_time,
            "bucket_size": bucket_size,
            "quantile": quantile,
            "threshold": threshold,
        }
        return ServiceStatsOverTime.get(self.id, **kwargs)

    def get_target_drift(self, model_id=None, start_time=None, end_time=None, metric=None):
        """Retrieve target drift information over a certain time period.

        .. versionadded:: v2.21

        Parameters
        ----------
        model_id : str
            the id of the model
        start_time : datetime
            start of the time period
        end_time : datetime
            end of the time period
        metric : str
            (New in version v2.22) metric used to calculate the drift score

        Returns
        -------
        target_drift : TargetDrift
            the queried target drift information
        """

        return TargetDrift.get(
            self.id, model_id=model_id, start_time=start_time, end_time=end_time, metric=metric
        )

    def get_feature_drift(self, model_id=None, start_time=None, end_time=None, metric=None):
        """Retrieve drift information for deployment's features over a certain time period.

        .. versionadded:: v2.21

        Parameters
        ----------
        model_id : str
            the id of the model
        start_time : datetime
            start of the time period
        end_time : datetime
            end of the time period
        metric : str
            (New in version v2.22) metric used to calculate the drift score

        Returns
        -------
        feature_drift_data : [FeatureDrift]
            the queried feature drift information
        """

        return FeatureDrift.list(
            self.id, model_id=model_id, start_time=start_time, end_time=end_time, metric=metric
        )

    def get_accuracy(self, model_id=None, start_time=None, end_time=None, start=None, end=None):
        """Retrieve values of accuracy metrics over a certain time period.

        .. versionadded:: v2.18

        Parameters
        ----------
        model_id : str
            the id of the model
        start_time : datetime
            start of the time period
        end_time : datetime
            end of the time period

        Returns
        -------
        accuracy : Accuracy
            the queried accuracy metrics information
        """
        # For a brief time, we accidentally used the kwargs "start" and "end". We add this logic
        # here to retain backwards compatibility with these legacy kwargs.
        start_time = start_time or start
        end_time = end_time or end

        return Accuracy.get(self.id, model_id=model_id, start_time=start_time, end_time=end_time)

    def get_accuracy_over_time(
        self, metric=None, model_id=None, start_time=None, end_time=None, bucket_size=None
    ):
        """Retrieve information about how an accuracy metric changes over a certain time period.

        .. versionadded:: v2.18

        Parameters
        ----------
        metric : ACCURACY_METRIC
            the accuracy metric to retrieve
        model_id : str
            the id of the model
        start_time : datetime
            start of the time period
        end_time : datetime
            end of the time period
        bucket_size : str
            time duration of a bucket, in ISO 8601 time duration format

        Returns
        -------
        accuracy_over_time : AccuracyOverTime
            the queried accuracy metric over time information
        """

        return AccuracyOverTime.get(
            self.id,
            metric=metric,
            model_id=model_id,
            start_time=start_time,
            end_time=end_time,
            bucket_size=bucket_size,
        )

    def update_secondary_dataset_config(self, secondary_dataset_config_id, credential_ids=None):
        """ Update the secondary dataset config used by Feature discovery model for a
        given deployment.

        .. versionadded:: v2.23

        Parameters
        ----------
        secondary_dataset_config_id: str
            Id of the secondary dataset config
        credential_ids: list or None
            List of DatasetsCredentials used by the secondary datasets

        Examples
        --------
        .. code-block:: python

            from datarobot import Deployment
            deployment = Deployment(deployment_id='5c939e08962d741e34f609f0')
            config = deployment.update_secondary_dataset_config('5df109112ca582033ff44084')
            config
            >>> '5df109112ca582033ff44084'
        """
        url = "{}{}/model/secondaryDatasetConfiguration/".format(self._path, self.id)
        payload = {"secondaryDatasetConfigId": secondary_dataset_config_id}
        if credential_ids:
            payload["credentialsIds"] = credential_ids
        self._client.patch(url, data=payload)
        return self.get_secondary_dataset_config()

    def get_secondary_dataset_config(self):
        """ Get the secondary dataset config used by Feature discovery model for a
        given deployment.

        .. versionadded:: v2.23

        Returns
        -------
        secondary_dataset_config : SecondaryDatasetConfigurations
            Id of the secondary dataset config

        Examples
        --------
        .. code-block:: python

            from datarobot import Deployment
            deployment = Deployment(deployment_id='5c939e08962d741e34f609f0')
            deployment.update_secondary_dataset_config('5df109112ca582033ff44084')
            config = deployment.get_secondary_dataset_config()
            config
            >>> '5df109112ca582033ff44084'
        """
        url = "{}{}/model/secondaryDatasetConfiguration/".format(self._path, self.id)
        response_json = from_api(self._client.get(url).json())
        return response_json.get("secondary_dataset_config_id")

    def get_prediction_results(
        self,
        model_id=None,
        start_time=None,
        end_time=None,
        actuals_present=None,
        offset=None,
        limit=None,
    ):
        """Retrieve a list of prediction results of the deployment.

        .. versionadded:: v2.24

        Parameters
        ----------
        model_id : str
            the id of the model
        start_time : datetime
            start of the time period
        end_time : datetime
            end of the time period
        actuals_present : bool
            filters predictions results to only those
            who have actuals present or with missing actuals
        offset : int
            this many results will be skipped
        limit : int
            at most this many results are returned

        Returns
        -------
        prediction_results: list[dict]
            a list of prediction results

        Examples
        --------
        .. code-block:: python

            from datarobot import Deployment
            deployment = Deployment.get(deployment_id='5c939e08962d741e34f609f0')
            results = deployment.get_prediction_results()
        """

        url = "{}{}/predictionResults/".format(self._path, self.id)
        params = self._build_query_params(
            start_time,
            end_time,
            model_id=model_id,
            actuals_present=actuals_present,
            offset=offset,
            limit=limit,
        )
        data = self._client.get(url, params=params).json()["data"]
        return from_api([prediction_result for prediction_result in data], keep_null_keys=True)

    def download_prediction_results(
        self,
        filepath,
        model_id=None,
        start_time=None,
        end_time=None,
        actuals_present=None,
        offset=None,
        limit=None,
    ):
        """Download prediction results of the deployment as a CSV file.

        .. versionadded:: v2.24

        Parameters
        ----------
        filepath : str
            path of the csv file
        model_id : str
            the id of the model
        start_time : datetime
            start of the time period
        end_time : datetime
            end of the time period
        actuals_present : bool
            filters predictions results to only those
            who have actuals present or with missing actuals
        offset : int
            this many results will be skipped
        limit : int
            at most this many results are returned

        Examples
        --------
        .. code-block:: python

            from datarobot import Deployment
            deployment = Deployment.get(deployment_id='5c939e08962d741e34f609f0')
            results = deployment.download_prediction_results('path_to_prediction_results.csv')
        """

        url = "{}{}/predictionResults/".format(self._path, self.id)
        headers = {"Accept": "text/csv"}
        params = self._build_query_params(
            start_time,
            end_time,
            model_id=model_id,
            actuals_present=actuals_present,
            offset=offset,
            limit=limit,
        )
        response = self._client.get(url, params=params, headers=headers)
        with open(filepath, mode="wb") as file:
            file.write(response.content)

    def download_scoring_code(self, filepath, source_code=False, include_agent=False):
        """Retrieve scoring code of the current deployed model.

        .. versionadded:: v2.24

        Notes
        -----
        When setting `include_agent` to `True`, it can take
        a considerably longer time to download the scoring code.

        Parameters
        ----------
        filepath : str
            path of the scoring code file
        source_code : bool
            whether source code or binary of the scoring code will be retrieved
        include_agent : bool
            whether the scoring code retrieved will include tracking agent

        Examples
        --------
        .. code-block:: python

            from datarobot import Deployment
            deployment = Deployment.get(deployment_id='5c939e08962d741e34f609f0')
            results = deployment.download_scoring_code('path_to_scoring_code.jar')
        """

        # retrieve the scoring code
        if include_agent:
            build_url = "{}{}/scoringCodeBuilds/".format(self._path, self.id)
            response = self._client.post(build_url, data={"includeAgent": include_agent})
            retrieve_url = wait_for_async_resolution(self._client, response.headers["Location"])
            response = self._client.get(retrieve_url)
        else:
            retrieve_url = "{}{}/scoringCode/".format(self._path, self.id)
            params = {"sourceCode": source_code, "includeAgent": include_agent}
            response = self._client.get(retrieve_url, params=params)

        # write to file
        with open(filepath, mode="wb") as file:
            file.write(response.content)


def _check(trafaret, default_to_null=True):
    def _trafaret_check(*args):
        value = args[0]

        if default_to_null and value is None:
            return None

        # Allow users to pass a single value even if a list of values is expected.
        if isinstance(trafaret, t.List) and not isinstance(value, list):
            value = [value]

        return trafaret.check(value)

    return _trafaret_check


@attr.s(slots=True, frozen=True)
class DeploymentListFilters(object):
    """Construct a set of filters to pass to ``Deployment.list()``

    .. versionadded:: v2.20

    Parameters
    ----------
    role : str
        A user role. If specified, then take those deployments that the user can view, then
        filter them down to those that the user has the specified role for, and return only them.
        Allowed options are ``OWNER`` and ``USER``.
    service_health : list of str
        A list of service health status values. If specified, then only deployments whose
        service health status is one of these will be returned. See
        ``datarobot.enums.DEPLOYMENT_SERVICE_HEALTH_STATUS`` for allowed values.
        Supports comma-separated lists.
    model_health : list of str
        A list of model health status values. If specified, then only deployments whose model
        health status is one of these will be returned. See
        ``datarobot.enums.DEPLOYMENT_MODEL_HEALTH_STATUS`` for allowed values.
        Supports comma-separated lists.
    accuracy_health : list of str
        A list of accuracy health status values. If specified, then only deployments whose
        accuracy health status is one of these will be returned. See
        ``datarobot.enums.DEPLOYMENT_ACCURACY_HEALTH_STATUS`` for allowed values.
        Supports comma-separated lists.
    execution_environment_type : list of str
        A list of strings representing the type of the deployments' execution environment.
        If provided, then only return those deployments whose execution environment type is
        one of those provided. See ``datarobot.enums.DEPLOYMENT_EXECUTION_ENVIRONMENT_TYPE``
        for allowed values. Supports comma-separated lists.
    importance : list of str
        A list of strings representing the deployments' "importance".
        If provided, then only return those deployments whose importance
        is one of those provided. See ``datarobot.enums.DEPLOYMENT_IMPORTANCE``
        for allowed values. Supports comma-separated lists. Note that Approval Workflows must be
        enabled for your account to use this filter, otherwise the API will return a 403.

    Examples
    --------
    Multiple filters can be combined in interesting ways to return very specific subsets of
    deployments.

    *Performing AND logic*

        Providing multiple different parameters will result in AND logic between them. For example,
        the following will return all deployments that I own whose service health status is failing.

        .. code-block:: python

            from datarobot import Deployment
            from datarobot.models.deployment import DeploymentListFilters
            from datarobot.enums import DEPLOYMENT_SERVICE_HEALTH
            filters = DeploymentListFilters(
                role='OWNER',
                service_health=[DEPLOYMENT_SERVICE_HEALTH.FAILING]
            )
            deployments = Deployment.list(filters=filters)

    **Performing OR logic**

        Some filters support comma-separated lists (and will say so if they do). Providing a
        comma-separated list of values to a single filter performs OR logic between those values.
        For example, the following will return all deployments whose service health is either
        ``warning`` OR ``failing``.

        .. code-block:: python

            from datarobot import Deployment
            from datarobot.models.deployment import DeploymentListFilters
            from datarobot.enums import DEPLOYMENT_SERVICE_HEALTH
            filters = DeploymentListFilters(
                service_health=[
                    DEPLOYMENT_SERVICE_HEALTH.WARNING,
                    DEPLOYMENT_SERVICE_HEALTH.FAILING,
                ]
            )
            deployments = Deployment.list(filters=filters)

    Performing OR logic across different filter types is not supported.

    .. note::

        In all cases, you may only retrieve deployments for which you have at least the USER role
        for. Deployments for which you are a CONSUMER of will not be returned, regardless of the
        filters applied.
    """

    role = attr.ib(default=None, converter=_check(String()))
    service_health = attr.ib(default=None, converter=_check(t.List[String()]))
    model_health = attr.ib(default=None, converter=_check(t.List[String()]))
    accuracy_health = attr.ib(default=None, converter=_check(t.List[String()]))
    execution_environment_type = attr.ib(default=None, converter=_check(t.List[String()]))
    importance = attr.ib(default=None, converter=_check(t.List[String()]))

    def construct_query_args(self):
        query_args = {}

        if self.role:
            query_args["role"] = self.role
        if self.service_health:
            query_args["serviceHealth"] = self._list_to_comma_separated_string(self.service_health)
        if self.model_health:
            query_args["modelHealth"] = self._list_to_comma_separated_string(self.model_health)
        if self.accuracy_health:
            query_args["accuracyHealth"] = self._list_to_comma_separated_string(
                self.accuracy_health
            )
        if self.execution_environment_type:
            query_args["executionEnvironmentType"] = self._list_to_comma_separated_string(
                self.execution_environment_type
            )
        if self.importance:
            query_args["importance"] = self._list_to_comma_separated_string(self.importance)

        return query_args

    @staticmethod
    def _list_to_comma_separated_string(input_list):
        output_string = ""
        for list_item in input_list:
            output_string += "{},".format(list_item)
        output_string = output_string[:-1]
        return output_string



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from collections import defaultdict
import json

from datarobot.utils import camelize


def assert_version(version, version_json):
    assert version.id == version_json["id"]
    assert version.environment_id == version_json["environmentId"]
    assert version.label == version_json["label"]
    assert version.description == version_json["description"]
    assert version.build_status == version_json["buildStatus"]
    assert version.created_at == version_json["created"]
    assert version.docker_context_size == version_json.get("dockerContextSize")
    assert version.docker_image_size == version_json.get("dockerImageSize")


def assert_custom_task(task_obj, task_json):
    non_optional = [
        "id",
        "target_type",
        "created_at",
        "updated_at",
        "name",
        "description",
        "language",
        "created_by",
    ]
    for obj_key in non_optional:
        assert_key(task_obj, task_json, obj_key)

    optional = [
        "calibrate_predictions",
    ]
    for obj_key in optional:
        assert_optional_key(task_obj, task_json, obj_key)

    version_obj = task_obj.latest_version
    version_json = task_json["latestVersion"]
    if version_obj is None:
        assert version_json is None, (version_obj, version_json)
    else:
        assert_custom_task_version(version_obj, version_json)


def assert_custom_model_version(version, version_json):
    model_version_only_keys = (
        "custom_model_id",
        "network_egress_policy",
        "maximum_memory",
        "replicas",
    )
    for key in model_version_only_keys:
        assert_key(version, version_json, key)

    _assert_custom_version_common_fields(version, version_json)


def assert_custom_task_version(version, version_json):
    assert_key(version, version_json, "custom_task_id")

    _assert_custom_version_common_fields(version, version_json)

    obj_arguments = version.arguments
    json_arguments = version_json.get("arguments", [])
    assert_custom_version_arguments(obj_arguments, json_arguments)


def _assert_custom_version_common_fields(version, version_json):
    non_optional = [
        "id",
        "label",
        "description",
        "version_minor",
        "version_major",
        "is_frozen",
        "created_at",
    ]
    for obj_key in non_optional:
        assert_key(version, version_json, obj_key)

    optional = [
        "base_environment_id",
        "required_metadata",
        "base_environment_version_id",
    ]
    for obj_key in optional:
        assert_optional_key(version, version_json, obj_key)

    if version.base_environment_id is not None:
        assert version.base_environment_version_id is not None

    object_items = version.items
    json_items = version_json["items"]
    assert_items(object_items, json_items)

    object_values = version.required_metadata_values or []
    json_values = version_json.get("requiredMetadataValues", [])
    assert_required_metadata_values(object_values, json_values)

    obj_dependencies = version.dependencies
    json_dependencies = version_json.get("dependencies", [])
    assert_dependencies(obj_dependencies, json_dependencies)


def assert_items(object_items, json_items):
    assert len(object_items) == len(json_items)
    for item, item_json in zip(object_items, json_items):
        for key in ("id", "file_name", "file_path", "file_source", "created_at"):
            assert_key(item, item_json, key)


def assert_required_metadata_values(object_values, json_values):
    assert len(object_values) == len(json_values)
    for item, item_json in zip(object_values, json_values):
        for key in ("field_name", "value"):
            assert_key(item, item_json, key)


def assert_dependencies(obj_dependencies, json_dependencies):
    assert len(obj_dependencies) == len(json_dependencies)
    for dependency, dependency_json in zip(obj_dependencies, json_dependencies):
        for key in ("package_name", "line", "line_number"):
            assert_key(dependency, dependency_json, key)

        for constraint, constraint_json in zip(
            dependency.constraints, dependency_json["constraints"]
        ):
            for key in ("version", "constraint_type"):
                assert_key(constraint, constraint_json, key)


def assert_custom_version_arguments(obj_arguments, json_arguments):
    assert len(obj_arguments) == len(json_arguments)
    for argument, argument_json in zip(obj_arguments, json_arguments):
        assert argument.key == argument_json["key"]
        for key in {"name", "type", "default", "values"}:
            assert_key(argument.argument, argument_json["argument"], key)


def assert_key(api_obj, json_data, obj_attribute):
    is_optional_key = False
    _assert_attribute(api_obj, json_data, obj_attribute, is_optional_key)


def assert_optional_key(api_obj, json_data, obj_attribute):
    is_optional_key = True
    _assert_attribute(api_obj, json_data, obj_attribute, is_optional_key)


def _assert_attribute(api_obj, json_data, obj_attribute, is_json_value_optional):
    json_attribute = camelize(obj_attribute)
    if obj_attribute in ("created_at", "updated_at"):
        json_attribute = obj_attribute.replace("_at", "")

    if is_json_value_optional:
        json_value = json_data.get(json_attribute)
    else:
        json_value = json_data[json_attribute]
    obj_value = getattr(api_obj, obj_attribute)

    assert obj_value == json_value, (obj_attribute, obj_value, json_value)


def assert_custom_model_version_dependency_build(
    build_info, build_info_json, custom_model_id, custom_model_version_id
):
    assert build_info.custom_model_id == custom_model_id
    assert build_info.custom_model_version_id == custom_model_version_id

    assert build_info.started_at == build_info_json["buildStart"]
    assert build_info.completed_at == build_info_json["buildEnd"]
    assert build_info.build_status == build_info_json["buildStatus"]


def change_response_value(value):
    """When you want to copy a response dict but change the values"""
    if isinstance(value, bool):
        return not value
    elif isinstance(value, int):
        return value + 1
    elif isinstance(value, str):
        return value + "x"
    elif isinstance(value, list):
        return value + value
    elif isinstance(value, dict):
        return {k + "x": v for k, v in value.items()}
    else:
        return value


def assert_single_api_call(calls, action, url_ending, request_body=None):
    # type: (responses.CallList, str, str, dict) -> None
    assert len(calls) == 1
    only_request = calls[0].request
    assert only_request.method == action
    assert only_request.url.endswith(url_ending)
    if request_body:
        actual_body = json.loads(only_request.body)
        assert actual_body == request_body, actual_body


def fields_to_dict(encoder):
    # type: (MultipartEncoder) -> defaultdict
    fields = defaultdict(list)
    for k, v in encoder.fields:
        fields[k].append(v)
    return fields


def assert_custom_task_version_dependency_build(
    custom_task_version_dependency_build, custom_task_version_dependency_build_server_data
):
    server_data = custom_task_version_dependency_build_server_data
    assert custom_task_version_dependency_build.custom_task_id == server_data["customTaskId"]
    assert (
        custom_task_version_dependency_build.custom_task_version_id
        == server_data["customTaskVersionId"]
    )
    assert custom_task_version_dependency_build.started_at == server_data["buildStart"]
    assert custom_task_version_dependency_build.completed_at == server_data["buildEnd"]
    assert custom_task_version_dependency_build.build_status == server_data["buildStatus"]
    assert (
        custom_task_version_dependency_build.build_log_location == server_data["buildLogLocation"]
    )



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json

import pytest
import responses
import six
import six.moves.urllib.parse as urllibparse

from datarobot import errors, UserBlueprint
from datarobot.models.user_blueprints.models import (
    DEFAULT_BATCH_SIZE,
    GrantAccessControlWithUsernameValidator,
    UserBlueprintAddToProjectMenu,
    UserBlueprintAvailableInput,
    UserBlueprintAvailableTasks,
    UserBlueprintCatalogSearch,
    UserBlueprintSharingListController,
    UserBlueprintValidate,
    UserBlueprintValidateTaskParameters,
    VertexContextItemMessages,
)
from datarobot.utils import from_api, to_api


def as_shallow_dict(item):
    return {
        k: v for k, v in six.iteritems(item.__dict__) if not k.startswith("_") and not callable(v)
    }


@pytest.fixture
def mock_role():
    return {
        "share_recipient_type": "user",
        "role": "OWNER",
        "id": "".join(["a"] * 24),
        "name": "test-user@datarobot.com",
    }


@pytest.fixture
def mock_roles_response(mock_role):
    return {"count": 1, "next": None, "previous": None, "data": [mock_role], "total_count": 1}


def build_mocked_task(name):
    return {
        "taskCode": name,
        "taskDefinition": {
            "customTaskId": None,
            "taskCode": "ABC",
            "description": (
                "AdaBoost Classifier (scikit-learn). An AdaBoost classifier "
                "is a meta-estimator that begins by fitting a classifier on the "
                "original dataset and then fits additional copies of the classifier "
                "on the same dataset but where the weights of incorrectly classified "
                "instances are adjusted such that subsequent classifiers focus more on "
                "difficult cases."
            ),
            "url": {"DataRobot Model Docs": "/model-docs/tasks/ABC-Adaboost-Classifier.html"},
            "sparseInputSupportLevel": "forbidden",
            "isCustomTask": None,
            "sparseInput": False,
            "categories": ["Adaptive Boosting Classifier"],
            "validInputs": ["NUM"],
            "label": "Adaboost Classifier",
            "colnamesAndTypes": [
                {"colname": "a", "hex": "61", "type": "NUM"},
                {"colname": "b", "hex": "62", "type": "NUM"},
            ],
            "arguments": [
                {
                    "argument": {
                        "default": "1234",
                        "values": [0, 1000000000],
                        "recommended": 1234,
                        "type": "int",
                        "name": "random_state",
                    },
                    "key": "rs",
                },
            ],
            "outputMethods": ["P", "S"],
            "hasSparseOutput": "unknown",
            "timeSeriesOnly": False,
            "outputsNonnumericInputType": False,
            "icon": 1,
            "customTaskVersions": [
                {"id": "abc", "versionMajor": -1, "versionMinor": -1, "label": "A"}
            ],
        },
    }


@pytest.fixture
def mocked_available_tasks():
    task = build_mocked_task("DEF")
    task["taskDefinition"]["supportsScoringCode"] = False
    return {
        "tasks": [build_mocked_task("ABC"), task],
        "categories": [
            {
                "name": "Adaptive Boosting Classifier",
                "taskCodes": ["ABC"],
                "subcategories": [{"name": "Adaptive Boosting Classifier", "taskCodes": ["ABC"]}],
            }
        ],
    }


@pytest.fixture
def mocked_validated_task_parameters():
    return {
        "errors": [
            {"paramName": "learning_rate", "message": "Invalid value(s) supplied", "value": "abc"},
        ],
    }


@pytest.fixture
def mocked_input_types():
    long_name_lookup = {
        "NUM": "Numeric",
        "CAT": "Categorical",
        "TXT": "Text",
    }
    return {
        "inputTypes": sorted(
            [{"type": _type, "name": name} for _type, name in six.iteritems(long_name_lookup)],
            key=lambda x: x["type"],
        )
    }


@pytest.fixture
def mocked_add_to_menu_response(blueprint_id, user_blueprint_id):
    return {
        "addedToMenu": [{"userBlueprintId": user_blueprint_id, "blueprintId": blueprint_id}],
        "notAddedToMenu": [],
        "message": "All blueprints successfully added to project repository.",
    }


@pytest.fixture
def blueprint():
    return [
        {
            "task_data": {
                "inputs": ["CAT"],
                "output_method": "T",
                "task_code": "ORDCAT2",
                "output_method_parameters": [],
                "x_transformations": [],
                "task_parameters": [
                    {"value": False, "param": "acm"},
                    {"value": False, "param": "amm"},
                    {"value": 5, "param": "ms"},
                    {"value": False, "param": "np_dtype_fix"},
                    {"value": True, "param": "o"},
                    {"value": 0, "param": "os"},
                    {"value": True, "param": "r"},
                    {"value": 1234, "param": "s"},
                ],
                "y_transformations": [],
            },
            "task_id": u"1",
        },
        {
            "task_data": {
                "inputs": ["NUM"],
                "output_method": "T",
                "task_code": "PNI2",
                "output_method_parameters": [],
                "x_transformations": [],
                "task_parameters": [
                    {"value": False, "param": "np_dtype_fix"},
                    {"value": False, "param": "s"},
                    {"value": 10, "param": "t"},
                ],
                "y_transformations": [],
            },
            "task_id": u"2",
        },
        {
            "task_data": {
                "inputs": [u"1", u"2"],
                "output_method": "P",
                "task_code": u"ESXGBC2",
                "output_method_parameters": [],
                "x_transformations": [],
                "task_parameters": [
                    {"value": 1, "param": "b_n"},
                    {"value": 1, "param": "b_nj"},
                    {"value": 1234, "param": "b_rs"},
                    {"value": False, "param": "base_init"},
                    {"value": 1.0, "param": "cbl"},
                    {"value": 1.0, "param": "cbt"},
                    {"value": u"deviance", "param": "l"},
                    {"value": None, "param": u"logy"},
                    {"value": 0.5, "param": "lr"},
                    {"value": 256, "param": "mb"},
                    {"value": 1.0, "param": "mcw"},
                    {"value": 3, "param": "md"},
                    {"value": 0.0, "param": "mds"},
                    {"value": 0.01, "param": "msl"},
                    {"value": 0.0, "param": "mv"},
                    {"value": 100, "param": "n"},
                    {"value": 1, "param": "npt"},
                    {"value": False, "param": "pp_mi"},
                    {"value": 5, "param": "pp_sf"},
                    {"value": True, "param": "pp_stk"},
                    {"value": 0, "param": "pp_stk_keep_top_n"},
                    {"value": False, "param": "pp_stk_seq"},
                    {"value": False, "param": "pp_stkm"},
                    {"value": False, "param": "pp_toboost"},
                    {"value": False, "param": "pp_wc"},
                    {"value": 0.0, "param": "ra"},
                    {"value": 1.0, "param": "rl"},
                    {"value": 1234, "param": "rs"},
                    {"value": 1.0, "param": "s"},
                    {"value": False, "param": "shap_center"},
                    {"value": False, "param": "shap_fit"},
                    {"value": 1.0, "param": "spw"},
                    {"value": 0, "param": "sw_b"},
                    {"value": 0.25, "param": "sw_tf"},
                    {"value": u"Pattern Search", "param": "t_a"},
                    {"value": 1, "param": "t_bn"},
                    {"value": True, "param": "t_call_fit_grid"},
                    {"value": False, "param": "t_clustering_score_on_training"},
                    {"value": True, "param": "t_es"},
                    {"value": 5, "param": "t_es_init"},
                    {"value": 10, "param": "t_int"},
                    {"value": False, "param": "t_light"},
                    {"value": 15, "param": "t_mi"},
                    {"value": 5, "param": "t_n"},
                    {"value": 1234, "param": "t_rs"},
                    {"value": True, "param": "t_s"},
                    {"value": False, "param": "t_sc"},
                    {"value": False, "param": "t_sf"},
                    {"value": 200, "param": "t_sint"},
                    {"value": 10, "param": "t_sp"},
                    {"value": u"auto", "param": "tm"},
                    {"value": False, "param": "wa_bw"},
                ],
                "y_transformations": [{"value": None, "param": u"logy"}],
            },
            "task_id": u"3",
        },
    ]


@pytest.fixture
def mocked_vertex_context(mock_user_blueprint_without_blueprint):
    return {
        "vertexContext": mock_user_blueprint_without_blueprint["vertexContext"],
        "blueprintContext": mock_user_blueprint_without_blueprint["blueprintContext"],
    }


@pytest.fixture
def blueprint_id():
    return "9c0c01ad4b23f1b055813ceb730bafca"


@pytest.fixture
def user_id():
    return "60733a301445e1d1625ebec5"


@pytest.fixture
def project_id():
    return "60733d631445e1d1625ebec6"


@pytest.fixture
def user_blueprint_id():
    return "556cdfbb100d2b0e88585182"


@pytest.fixture
def mock_user_blueprint_without_blueprint(blueprint_id, user_blueprint_id, user_id):
    return {
        "blender": False,
        "blueprintId": blueprint_id,
        "diagram": "{}",
        "features": ["Keras Neural Network Classifier"],
        "featuresText": "Keras Neural Network Classifier",
        "icons": [4],
        "insights": "NA",
        "modelType": "Keras Neural Network Classifier",
        "referenceModel": False,
        "supportsGpu": True,
        "shapSupport": False,
        "supportsNewSeries": False,
        "userBlueprintId": user_blueprint_id,
        "userId": str(user_id),
        "blueprintContext": {"warnings": [], "errors": []},
        "vertexContext": [
            {
                "information": {
                    "inputs": [
                        "Missing Values: Forbidden",
                        "Data Type: Numeric",
                        "Sparsity: Supported",
                    ],
                    "outputs": ["Missing Values: Never", "Data Type: Numeric", "Sparsity: Never"],
                },
                "messages": {
                    "warnings": ["Unexpected input type. Expected Numeric, received All."]
                },
                "taskId": "1",
            }
        ],
        "supportedTargetTypes": ["binary"],
        "isTimeSeries": False,
        "hexColumnNameLookup": [],
        "customTaskVersionMetadata": [],
        "decompressedFormat": False,
    }


@pytest.fixture
def mock_user_blueprint(blueprint, mock_user_blueprint_without_blueprint):
    return dict(mock_user_blueprint_without_blueprint, blueprint=from_api(blueprint))


@pytest.fixture
def mock_user_blueprint_with_project_id(mock_user_blueprint, project_id):
    return dict(mock_user_blueprint, projectId=project_id)


@pytest.fixture
def mock_user_blueprint_with_custom_task_version(mock_user_blueprint_without_blueprint):
    custom_task_version_id = "1" * 24
    custom_task_id = "2" * 24
    bp_with_custom_task = [
        {
            "task_id": "1",
            "task_data": {
                "inputs": ["CAT", "DATE", "DATE_DURATION", "NUM", "TXT"],
                "task_code": "CUSTOMR",
                "task_parameters": [{"param": "version_id", "value": custom_task_version_id}],
                "output_method": "P",
                "output_method_parameters": [],
                "x_transformations": [],
                "y_transformations": [],
            },
        },
    ]
    return dict(
        mock_user_blueprint_without_blueprint,
        blueprint=bp_with_custom_task,
        customTaskVersionMetadata=[[custom_task_version_id, custom_task_id]],
        modelType="training Model (v1.0)",
        taskType="custom",
    )


@pytest.fixture
def make_url(unittest_endpoint):
    def _make_url(url):
        return "{}/{}".format(unittest_endpoint, url)

    return _make_url


@pytest.mark.parametrize("limit", [0, 10])
@responses.activate
def test_list_blueprints(make_url, mock_user_blueprint_without_blueprint, limit):
    original_limit = limit
    if limit == 0:
        limit = DEFAULT_BATCH_SIZE
    urlA = make_url("userBlueprints/?offset=3&limit={}").format(limit)
    urlB = make_url("userBlueprints/?limit={}&offset=3").format(limit)
    responses.add(
        responses.GET,
        urlA,
        status=200,
        content_type="application/json",
        body=json.dumps({"data": [mock_user_blueprint_without_blueprint], "next": None}),
    )
    response = UserBlueprint.list(limit=original_limit, offset=3)
    assert response == [UserBlueprint.from_server_data(mock_user_blueprint_without_blueprint)]
    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url in [urlA, urlB]


@pytest.mark.parametrize("project_id, expected_project_id", [(None, None), ("foo", "foo")])
@responses.activate
def test_get_blueprint(
    project_id, expected_project_id, make_url, user_blueprint_id, mock_user_blueprint
):
    if project_id is None:
        urlA = make_url(
            "userBlueprints/{}/?editMode=True&decompressedBlueprint=True".format(user_blueprint_id)
        )
        urlB = make_url(
            "userBlueprints/{}/?decompressedBlueprint=True&editMode=True".format(user_blueprint_id)
        )
    else:
        url = (
            "userBlueprints/{user_bp_id}/?editMode=True&decompressedBlueprint=True&projectId={pid}"
        )
        urlA = make_url(url.format(user_bp_id=user_blueprint_id, pid=expected_project_id))
        url = (
            "userBlueprints/{user_bp_id}/?projectId={pid}&editMode=True&decompressedBlueprint=True"
        )
        urlB = make_url(url.format(user_bp_id=user_blueprint_id, pid=expected_project_id))
    responses.add(
        responses.GET,
        urlA,
        status=200,
        content_type="application/json",
        body=json.dumps(mock_user_blueprint),
    )
    response = UserBlueprint.get(user_blueprint_id, project_id=project_id)
    assert response.blueprint_context == VertexContextItemMessages(errors=[], warnings=[],)
    assert response == UserBlueprint.from_server_data(mock_user_blueprint)
    assert response == UserBlueprint(**as_shallow_dict(response))
    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url in [urlA, urlB]


@responses.activate
@pytest.mark.parametrize("pid, expected_pid", [(None, None), ("foo", "foo")])
@pytest.mark.parametrize("model_type, expected_model_type", [(None, None), ("foo", "foo")])
@pytest.mark.parametrize(
    "save_to_catalog, expected_save_to_catalog", [(None, True), (False, False)]
)
def test_create_blueprint_with_data(
    make_url,
    mock_user_blueprint,
    blueprint,
    pid,
    expected_pid,
    model_type,
    expected_model_type,
    save_to_catalog,
    expected_save_to_catalog,
):
    url = make_url("userBlueprints/")
    responses.add(
        responses.POST,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(mock_user_blueprint),
    )

    kwargs = {}
    if pid is not None:
        kwargs.update(dict(project_id=pid))
    if model_type is not None:
        kwargs.update(dict(model_type=model_type))
    if save_to_catalog is not None:
        kwargs.update(dict(save_to_catalog=save_to_catalog))

    response = UserBlueprint.create(blueprint, **kwargs)
    assert response == UserBlueprint.from_server_data(mock_user_blueprint)
    assert response == UserBlueprint(**as_shallow_dict(response))
    assert responses.calls[0].request.method == "POST"
    assert responses.calls[0].request.url == url

    # Inspect and verify the request
    body = from_api(json.loads(responses.calls[0].request.body), keep_null_keys=True)
    expected = {
        "blueprint": blueprint,
        "decompressed_blueprint": True,
    }

    if expected_pid is not None:
        expected.update(dict(project_id=expected_pid))
    if expected_model_type is not None:
        expected.update(dict(model_type=expected_model_type))
    if expected_save_to_catalog is not None:
        expected.update(dict(save_to_catalog=expected_save_to_catalog))

    assert body == expected


@responses.activate
@pytest.mark.parametrize("model_type, expected_model_type", [(None, None), ("foo", "foo")])
@pytest.mark.parametrize(
    "save_to_catalog, expected_save_to_catalog", [(None, True), (False, False)]
)
def test_create_blueprint_with_blueprint_id_and_project_id(
    make_url,
    mock_user_blueprint_with_project_id,
    blueprint_id,
    project_id,
    model_type,
    expected_model_type,
    save_to_catalog,
    expected_save_to_catalog,
):
    url = make_url("userBlueprints/fromBlueprintId/")
    responses.add(
        responses.POST,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(mock_user_blueprint_with_project_id),
    )

    kwargs = {}
    if model_type is not None:
        kwargs.update(dict(model_type=model_type))
    if save_to_catalog is not None:
        kwargs.update(dict(save_to_catalog=save_to_catalog))

    response = UserBlueprint.clone_project_blueprint(blueprint_id, project_id, **kwargs)
    assert response == UserBlueprint.from_server_data(mock_user_blueprint_with_project_id)
    assert response == UserBlueprint(**as_shallow_dict(response))
    assert responses.calls[0].request.method == "POST"
    assert responses.calls[0].request.url == url

    # Inspect and verify the request
    body = from_api(json.loads(responses.calls[0].request.body), keep_null_keys=True)
    expected = {
        "blueprint_id": blueprint_id,
        "decompressed_blueprint": True,
        "project_id": project_id,
    }

    if expected_model_type is not None:
        expected.update(dict(model_type=expected_model_type))
    if expected_save_to_catalog is not None:
        expected.update(dict(save_to_catalog=expected_save_to_catalog))

    assert body == expected


@responses.activate
@pytest.mark.parametrize("pid, expected_pid", [(None, None), ("foo", "foo")])
@pytest.mark.parametrize("model_type, expected_model_type", [(None, None), ("foo", "foo")])
@pytest.mark.parametrize(
    "save_to_catalog, expected_save_to_catalog", [(None, True), (False, False)]
)
def test_create_blueprint_with_user_blueprint_id(
    make_url,
    mock_user_blueprint,
    user_blueprint_id,
    pid,
    expected_pid,
    model_type,
    expected_model_type,
    save_to_catalog,
    expected_save_to_catalog,
):
    url = make_url("userBlueprints/fromUserBlueprintId/")
    responses.add(
        responses.POST,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(mock_user_blueprint),
    )

    kwargs = {}
    if pid is not None:
        kwargs.update(dict(project_id=pid))
    if model_type is not None:
        kwargs.update(dict(model_type=model_type))
    if save_to_catalog is not None:
        kwargs.update(dict(save_to_catalog=save_to_catalog))

    response = UserBlueprint.clone_user_blueprint(user_blueprint_id, **kwargs)
    assert response == UserBlueprint.from_server_data(mock_user_blueprint)
    assert response == UserBlueprint(**as_shallow_dict(response))
    assert responses.calls[0].request.method == "POST"
    assert responses.calls[0].request.url == url

    # Inspect and verify the request
    body = from_api(json.loads(responses.calls[0].request.body), keep_null_keys=True)
    expected = {
        "user_blueprint_id": user_blueprint_id,
        "decompressed_blueprint": True,
    }

    if expected_pid is not None:
        expected.update(dict(project_id=expected_pid))
    if expected_model_type is not None:
        expected.update(dict(model_type=expected_model_type))
    if expected_save_to_catalog is not None:
        expected.update(dict(save_to_catalog=expected_save_to_catalog))

    assert body == expected


@pytest.mark.parametrize("set_blueprint_to_none", [False, True])
@pytest.mark.parametrize("project_id", ["foo", None])
@pytest.mark.parametrize("include_project_id_if_none", [False, True])
@responses.activate
def test_update_blueprint_with_data(
    make_url,
    user_blueprint_id,
    mock_user_blueprint,
    blueprint,
    set_blueprint_to_none,
    project_id,
    include_project_id_if_none,
):
    url = make_url("userBlueprints/{}/".format(user_blueprint_id))
    responses.add(
        responses.PATCH,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(mock_user_blueprint),
    )
    model_type = "Foo Bar"
    if set_blueprint_to_none:
        blueprint = None
    response = UserBlueprint.update(
        blueprint,
        user_blueprint_id,
        model_type=model_type,
        project_id=project_id,
        include_project_id_if_none=include_project_id_if_none,
    )
    assert response == UserBlueprint.from_server_data(mock_user_blueprint)
    assert response == UserBlueprint(**as_shallow_dict(response))
    assert responses.calls[0].request.method == "PATCH"
    assert responses.calls[0].request.url == url
    expected_body = {"decompressed_blueprint": True, "model_type": "Foo Bar"}
    if project_id is not None or include_project_id_if_none:
        expected_body["project_id"] = project_id
    if not set_blueprint_to_none:
        expected_body["blueprint"] = blueprint
    assert (
        from_api(json.loads(responses.calls[0].request.body), keep_null_keys=True) == expected_body
    )


@responses.activate
def test_delete_blueprint_with_data(make_url, user_blueprint_id):
    url = make_url("userBlueprints/{}/".format(user_blueprint_id))
    responses.add(
        responses.DELETE, url, status=204, content_type="application/json",
    )
    UserBlueprint.delete(user_blueprint_id)
    assert responses.calls[0].request.method == "DELETE"
    assert responses.calls[0].request.url == url


@responses.activate
def test_add_to_project(
    make_url, project_id, user_blueprint_id, blueprint_id, mocked_add_to_menu_response,
):
    url = make_url("userBlueprintsProjectBlueprints/")
    responses.add(
        responses.POST,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(mocked_add_to_menu_response),
    )

    response = UserBlueprint.add_to_project(
        project_id=project_id, user_blueprint_ids=[user_blueprint_id]
    )
    assert response == UserBlueprintAddToProjectMenu(**as_shallow_dict(response))
    assert len(response.added_to_menu) == 1
    assert response.added_to_menu[0].blueprint_id == blueprint_id
    assert response.added_to_menu[0].user_blueprint_id == user_blueprint_id
    assert response.not_added_to_menu == []
    assert response.message == "All blueprints successfully added to project repository."
    assert responses.calls[0].request.method == "POST"
    assert responses.calls[0].request.url == url


@responses.activate
def test_get_input_types(
    make_url, project_id, user_blueprint_id, blueprint_id, mocked_input_types,
):
    url = make_url("userBlueprintsInputTypes/")
    responses.add(
        responses.GET,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(mocked_input_types),
    )

    response = UserBlueprint.get_input_types()
    assert response == UserBlueprintAvailableInput.from_server_data(mocked_input_types)
    assert response == UserBlueprintAvailableInput(**as_shallow_dict(response))
    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url


@responses.activate
def test_get_available_tasks(
    make_url, project_id, user_blueprint_id, blueprint_id, mocked_available_tasks,
):
    url = make_url("userBlueprintsTasks/")
    responses.add(
        responses.GET,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(mocked_available_tasks),
    )

    response = UserBlueprint.get_available_tasks()
    assert response.tasks[0].task_definition.url.documentation == (
        "/model-docs/tasks/ABC-Adaboost-Classifier.html"
    )
    assert response == UserBlueprintAvailableTasks.from_server_data(mocked_available_tasks)
    assert response == UserBlueprintAvailableTasks(**as_shallow_dict(response))
    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url


@pytest.mark.parametrize("project_id, expected_project_id", [(None, None), ("foo", "foo")])
@responses.activate
def test_validate_task_parameters(
    project_id,
    expected_project_id,
    make_url,
    user_blueprint_id,
    blueprint_id,
    mocked_validated_task_parameters,
):
    url = make_url("userBlueprintsTaskParameters/")
    responses.add(
        responses.POST,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(mocked_validated_task_parameters),
    )
    kwargs = {
        "output_method": "P",
        "task_code": "KERASC",
        "task_parameters": [{"paramName": "learning_rate", "newValue": "abc"}],
    }
    if project_id is not None:
        kwargs.update(dict(project_id=project_id))
    response = UserBlueprint.validate_task_parameters(**kwargs)

    # Inspect and verify the request
    request_body = json.loads(responses.calls[0].request.body.decode("utf-8").replace("'", '"'))
    expected_request = {
        "output_method": "P",
        "task_code": "KERASC",
        "task_parameters": [{"paramName": "learning_rate", "newValue": "abc"}],
    }
    if project_id is not None:
        expected_request.update(dict(project_id=project_id))
    assert request_body == to_api(expected_request)

    # Inspect response
    assert response == (
        UserBlueprintValidateTaskParameters.from_server_data(mocked_validated_task_parameters)
    )
    assert response == UserBlueprintValidateTaskParameters(**as_shallow_dict(response))
    assert responses.calls[0].request.method == "POST"
    assert responses.calls[0].request.url == url


@pytest.mark.parametrize("project_id, expected_project_id", [(None, None), ("foo", "foo")])
@responses.activate
def test_validate_blueprint(
    project_id, expected_project_id, make_url, blueprint, mocked_vertex_context,
):
    url = make_url("userBlueprintsValidations/")
    responses.add(
        responses.POST,
        url,
        status=201,
        content_type="application/json",
        body=json.dumps(mocked_vertex_context),
    )
    kwargs = {
        "blueprint": blueprint,
    }
    if project_id is not None:
        kwargs.update(dict(project_id=project_id))
    response = UserBlueprint.validate_blueprint(**kwargs)
    assert response.blueprint_context == VertexContextItemMessages(errors=[], warnings=[],)

    # Inspect and verify the request
    request_body = json.loads(responses.calls[0].request.body.decode("utf-8").replace("'", '"'))
    expected_request = {
        "blueprint": blueprint,
    }
    if project_id is not None:
        expected_request.update(dict(project_id=expected_project_id))
    assert request_body == to_api(expected_request)

    # Inspect response
    assert response == (UserBlueprintValidate.from_server_data(mocked_vertex_context))
    assert response == UserBlueprintValidate(**as_shallow_dict(response))
    assert responses.calls[0].request.method == "POST"
    assert responses.calls[0].request.url == url


@pytest.mark.parametrize("limit", [0, 10])
@responses.activate
def test_retrieve_shared_roles(make_url, user_blueprint_id, mock_roles_response, limit):
    original_limit = limit
    if limit == 0:
        limit = DEFAULT_BATCH_SIZE
    urlA = make_url("userBlueprints/{}/sharedRoles/?offset=3&limit={}").format(
        user_blueprint_id, limit
    )
    urlB = make_url("userBlueprints/{}/sharedRoles/?limit={}&offset=3").format(
        user_blueprint_id, limit
    )
    responses.add(
        responses.GET,
        urlA,
        status=200,
        content_type="application/json",
        body=json.dumps(mock_roles_response),
    )
    response = UserBlueprint.list_shared_roles(user_blueprint_id, limit=original_limit, offset=3)
    assert response == (
        UserBlueprintSharingListController.from_server_data(mock_roles_response).data
    )
    assert response == UserBlueprintSharingListController(data=response).data
    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url in [urlA, urlB]


@pytest.mark.parametrize("recipientType", ["user", "group", "organization"])
@pytest.mark.parametrize("role", ["CONSUMER", "EDITOR", "OWNER"])
@responses.activate
def test_update_shared_roles(make_url, user_blueprint_id, role, recipientType):
    url = make_url("userBlueprints/{}/sharedRoles/".format(user_blueprint_id))
    responses.add(
        responses.PATCH, url, status=204, content_type="application/json",
    )
    response = UserBlueprint.update_shared_roles(
        user_blueprint_id, [GrantAccessControlWithUsernameValidator(role, recipientType, "test")]
    )
    assert response.status_code == 204
    assert responses.calls[0].request.method == "PATCH"
    assert responses.calls[0].request.url == url


@responses.activate
def test_catalog_search(make_url):
    url = make_url("catalogItems/")
    responses.add(
        responses.GET,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(
            {
                "data": [
                    to_api(
                        dict(
                            id="a",
                            catalog_name="b",
                            info_creator_full_name="c",
                            user_blueprint_id="d",
                            description="e",
                            last_modifier_name="f",
                        )
                    )
                ]
            }
        ),
    )
    response = UserBlueprint.search_catalog()
    assert response == [
        UserBlueprintCatalogSearch(
            id="a",
            catalog_name="b",
            info_creator_full_name="c",
            user_blueprint_id="d",
            description="e",
            last_modifier_name="f",
        )
    ]
    assert responses.calls[0].response.status_code == 200
    assert responses.calls[0].request.method == "GET"
    request_url = responses.calls[0].request.url
    parsed_url = urllibparse.urlparse(request_url)
    qs = urllibparse.parse_qs(parsed_url.query)
    assert qs == dict(limit=["100"], offset=["0"], orderBy=["-created"], type=["user_blueprint"])

    response = UserBlueprint.search_catalog(
        search="foo",
        tag="a,b,c",
        limit=10,
        offset=5,
        owner_user_id="test_id",
        owner_username="test_name",
        order_by="relevance",
    )
    assert response == [
        UserBlueprintCatalogSearch(
            id="a",
            catalog_name="b",
            info_creator_full_name="c",
            user_blueprint_id="d",
            description="e",
            last_modifier_name="f",
        )
    ]
    assert responses.calls[1].response.status_code == 200
    assert responses.calls[1].request.method == "GET"
    request_url = responses.calls[1].request.url
    parsed_url = urllibparse.urlparse(request_url)
    qs = urllibparse.parse_qs(parsed_url.query)
    assert qs == dict(
        limit=["10"],
        offset=["5"],
        orderBy=["relevance"],
        ownerUserId=["test_id"],
        ownerUsername=["test_name"],
        searchFor=["foo"],
        tag=["a,b,c"],
        type=["user_blueprint"],
    )


@responses.activate
def test_create_blueprint_with_custom_task_version_id_req_param__default_value(
    make_url, mock_user_blueprint
):
    url = make_url("userBlueprints/fromCustomTaskVersionId/")
    responses.add(
        responses.POST,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(mock_user_blueprint),
    )
    # Create the user BP.
    custom_task_version_id = "1" * 24
    UserBlueprint.create_from_custom_task_version_id(custom_task_version_id=custom_task_version_id)
    # Check the public API is properly called.
    request_body = json.loads(responses.calls[0].request.body.decode("utf-8"))
    expected_request_body = {
        "customTaskVersionId": custom_task_version_id,
        "decompressedBlueprint": True,
        "saveToCatalog": True,
    }
    assert request_body == expected_request_body


@responses.activate
@pytest.mark.parametrize("custom_task_version_id", [1111, "1111"])
def test_create_blueprint_with_custom_task_version_id_req_param__custom_task_version_id(
    make_url, mock_user_blueprint, custom_task_version_id,
):
    url = make_url("userBlueprints/fromCustomTaskVersionId/")
    responses.add(
        responses.POST,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(mock_user_blueprint),
    )
    # Create the user BP.
    UserBlueprint.create_from_custom_task_version_id(custom_task_version_id=custom_task_version_id)
    # Check the public API is properly called.
    request_body = json.loads(responses.calls[0].request.body.decode("utf-8"))
    expected_request_body = {
        "customTaskVersionId": "1111",
        "decompressedBlueprint": True,
        "saveToCatalog": True,
    }
    assert request_body == expected_request_body


@responses.activate
@pytest.mark.parametrize("save_to_catalog", [True, False, None])
def test_create_blueprint_with_custom_task_version_id_req_param__save_to_catalog(
    make_url, mock_user_blueprint, save_to_catalog,
):
    url = make_url("userBlueprints/fromCustomTaskVersionId/")
    responses.add(
        responses.POST,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(mock_user_blueprint),
    )
    # Create the user BP.
    custom_task_version_id = "1" * 24
    kwargs = {}
    if save_to_catalog is not None:
        kwargs["save_to_catalog"] = save_to_catalog
    UserBlueprint.create_from_custom_task_version_id(
        custom_task_version_id=custom_task_version_id, **kwargs
    )
    # Check the public API is properly called.
    request_body = json.loads(responses.calls[0].request.body.decode("utf-8"))
    expected_request_body = {
        "customTaskVersionId": custom_task_version_id,
        "decompressedBlueprint": True,
        "saveToCatalog": save_to_catalog if save_to_catalog is not None else True,
    }
    assert request_body == expected_request_body


@responses.activate
@pytest.mark.parametrize("description", ["", "non blank", None])
def test_create_blueprint_with_custom_task_version_id_req_param__description(
    make_url, mock_user_blueprint, description,
):
    url = make_url("userBlueprints/fromCustomTaskVersionId/")
    responses.add(
        responses.POST,
        url,
        status=200,
        content_type="application/json",
        body=json.dumps(mock_user_blueprint),
    )
    # Create the user BP.
    custom_task_version_id = "1" * 24
    kwargs = {}
    if description is not None:
        kwargs["description"] = description
    # Check UserBlueprint is successfully created with public api response.
    UserBlueprint.create_from_custom_task_version_id(
        custom_task_version_id=custom_task_version_id, **kwargs
    )
    # Check the public API is properly called.
    request_body = json.loads(responses.calls[0].request.body.decode("utf-8"))
    expected_request_body = {
        "customTaskVersionId": custom_task_version_id,
        "saveToCatalog": True,
        "decompressedBlueprint": True,
    }
    if description is not None:
        expected_request_body.update({"description": description})
    assert request_body == expected_request_body


@responses.activate
@pytest.mark.parametrize("http_status_code", [404, 403, 422, 500])
def test_create_blueprint_with_custom_task_version_id__raise_proper_exception(
    make_url, http_status_code
):
    url = make_url("userBlueprints/fromCustomTaskVersionId/")
    responses.add(
        responses.POST,
        url,
        status=http_status_code,
        content_type="application/json",
        body=json.dumps({"message": "error message"}),
    )
    is_client_error = 400 <= http_status_code < 500
    expected_raised_error = errors.ClientError if is_client_error else errors.ServerError
    with pytest.raises(expected_raised_error) as error:
        UserBlueprint.create_from_custom_task_version_id(custom_task_version_id="1" * 24)
    error_obj = error.value
    assert error_obj.status_code == http_status_code


def test_from_server_data_with_custom_task_blueprint(mock_user_blueprint_without_blueprint):
    custom_task_version_id = "1" * 24
    custom_task_id = "2" * 24
    bp_with_custom_task = [
        {
            "task_id": "1",
            "task_data": {
                "inputs": ["CAT", "DATE", "DATE_DURATION", "NUM", "TXT"],
                "task_code": "CUSTOMR",
                "task_parameters": [{"param": "version_id", "value": custom_task_version_id}],
                "output_method": "P",
                "output_method_parameters": [],
                "x_transformations": [],
                "y_transformations": [],
            },
        },
    ]
    mock_user_blueprint_with_custom_task_version = dict(
        mock_user_blueprint_without_blueprint,
        blueprint=bp_with_custom_task,
        customTaskVersionMetadata=[[custom_task_version_id, custom_task_id]],
        modelType="training Model (v1.0)",
        taskType="custom",
    )
    user_blueprint = UserBlueprint.from_server_data(mock_user_blueprint_with_custom_task_version)
    assert user_blueprint.custom_task_version_metadata == [[custom_task_version_id, custom_task_id]]



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json

import pytest
import responses

from datarobot import DatetimeModel, Model, ModelRecommendation
from datarobot.enums import RECOMMENDED_MODEL_TYPE


@pytest.fixture
def list_recommended_model_json():
    model_list = [
        {
            "recommendationType": RECOMMENDED_MODEL_TYPE.MOST_ACCURATE,
            "modelId": "model1",
            "projectId": "project1",
        },
        {
            "recommendationType": RECOMMENDED_MODEL_TYPE.FAST_ACCURATE,
            "modelId": "model2",
            "projectId": "project1",
        },
    ]
    return json.dumps(model_list)


@pytest.fixture
def recommended_model_json():
    model = {
        "recommendationType": "Fast & Accurate",
        "modelId": "5ad8e121fcf4c87d96ae02dd",
        "projectId": "project1",
    }
    return json.dumps(model)


@pytest.fixture
def recommended_model_extra_field_json():
    model = {
        "recommendationType": "Fast & Accurate",
        "modelId": "5ad8e121fcf4c87d96ae02dd",
        "projectId": "project1",
        "extraField": "value",
    }
    return json.dumps(model)


@pytest.fixture
def non_datetime_project_json(project_id):
    project = {
        "partition": {"cvMethod": "random"},
        "projectId": project_id,
    }
    return json.dumps(project)


@pytest.fixture
def datetime_project_json(project_id):
    project = {
        "partition": {"cvMethod": "datetime"},
        "projectId": project_id,
    }
    return json.dumps(project)


@pytest.fixture
def most_accurate_recommended_model_json(project_id, model_id):
    model = {
        "recommendationType": RECOMMENDED_MODEL_TYPE.MOST_ACCURATE,
        "modelId": model_id,
        "projectId": project_id,
    }
    return json.dumps(model)


@pytest.fixture
def recommended_datetime_model_json(project_id, model_id):
    model = {
        "backtests": [
            {
                "index": 0,
                "score": 0.7,
                "status": "COMPLETED",
                "training_start_date": "2015-10-26T19:00:00.000000Z",
                "training_duration": "P0Y1M0D",
                "training_row_count": 880,
                "training_end_date": "2015-11-26T19:00:00.000000Z",
            },
        ],
        "projectId": project_id,
        "dataSelectionMethod": "rowCount",
        "trainingInfo": {
            "predictionTrainingDuration": "P0Y1M0D",
            "holdoutTrainingStartDate": "2015-10-26T19:00:00.000000Z",
            "predictionTrainingStartDate": "2015-10-26T19:00:00.000000Z",
            "holdoutTrainingDuration": "P0Y1M0D",
            "predictionTrainingRowCount": 880,
            "holdoutTrainingEndDate": "2015-12-10T19:00:00.000000Z",
            "predictionTrainingEndDate": "2015-12-10T19:00:00.000000Z",
            "holdoutTrainingRowCount": 880,
        },
        "id": model_id,
        "effective_feature_derivation_window_start": -120,
        "effective_feature_derivation_window_end": 0,
    }
    return json.dumps(model)


@responses.activate
def test_recommended_model_default(recommended_model_json):
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/recommendedModels/recommendedModel/",
        body=recommended_model_json,
    )
    recommended_model = ModelRecommendation.get("p-id")
    assert recommended_model.recommendation_type == "Fast & Accurate"
    assert recommended_model.model_id == "5ad8e121fcf4c87d96ae02dd"
    assert recommended_model.project_id == "project1"


@responses.activate
def test_recommended_model_specific(list_recommended_model_json):
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/recommendedModels/",
        body=list_recommended_model_json,
    )
    recommended_model = ModelRecommendation.get("p-id", RECOMMENDED_MODEL_TYPE.MOST_ACCURATE)
    assert recommended_model.recommendation_type == "Most Accurate"
    assert recommended_model.model_id == "model1"
    assert recommended_model.project_id == "project1"


@responses.activate
def test_recommended_model_ignore_extra_fields(recommended_model_extra_field_json):
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/recommendedModels/recommendedModel/",
        body=recommended_model_extra_field_json,
    )
    recommended_model = ModelRecommendation.get("p-id")
    assert recommended_model.recommendation_type == "Fast & Accurate"
    assert recommended_model.model_id == "5ad8e121fcf4c87d96ae02dd"
    assert recommended_model.project_id == "project1"
    assert not hasattr(recommended_model, "extraField")


@responses.activate
def test_recommended_model_list(list_recommended_model_json):
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/recommendedModels/",
        body=list_recommended_model_json,
    )
    recommended_models = ModelRecommendation.get_all("p-id")
    assert len(recommended_models) == 2
    expected_models = json.loads(list_recommended_model_json)

    expected_model1 = next(model for model in expected_models if model["modelId"] == "model1")
    expected_model2 = next(model for model in expected_models if model["modelId"] == "model2")
    actual_model1 = [model for model in recommended_models if model.model_id == "model1"]
    actual_model2 = [model for model in recommended_models if model.model_id == "model2"]

    assert len(actual_model1) == 1
    assert len(actual_model2) == 1
    actual_model1 = actual_model1[0]
    actual_model2 = actual_model2[0]

    assert actual_model1.recommendation_type == expected_model1["recommendationType"]
    assert actual_model2.recommendation_type == expected_model2["recommendationType"]
    assert actual_model1.project_id == expected_model1["projectId"]
    assert actual_model2.project_id == expected_model2["projectId"]

    assert (
        ModelRecommendation.get_recommendation(
            recommended_models, RECOMMENDED_MODEL_TYPE.MOST_ACCURATE
        )
        is not None
    )

    assert (
        ModelRecommendation.get_recommendation(
            recommended_models, RECOMMENDED_MODEL_TYPE.RECOMMENDED_FOR_DEPLOYMENT
        )
        is None
    )


@responses.activate
def test_recommended_non_datetime_model(
    project_id, model_id, non_datetime_project_json, most_accurate_recommended_model_json
):
    responses.add(
        responses.GET,
        "https://host_name.com/projects/{pid}/".format(pid=project_id),
        status=200,
        body=non_datetime_project_json,
    )
    responses.add(
        responses.GET,
        "https://host_name.com/projects/{pid}/models/{lid}/".format(pid=project_id, lid=model_id),
        status=200,
        body=most_accurate_recommended_model_json,
    )
    recommendation = ModelRecommendation(project_id, model_id, RECOMMENDED_MODEL_TYPE.MOST_ACCURATE)
    model = recommendation.get_model()
    assert isinstance(model, Model)
    assert hasattr(model, "model_type")
    assert not hasattr(model, "effective_feature_derivation_window_start")


@responses.activate
def test_recommended_datetime_model(
    project_id, model_id, datetime_project_json, recommended_datetime_model_json
):
    responses.add(
        responses.GET,
        "https://host_name.com/projects/{pid}/".format(pid=project_id),
        status=200,
        body=datetime_project_json,
    )
    responses.add(
        responses.GET,
        "https://host_name.com/projects/{pid}/datetimeModels/{lid}/".format(
            pid=project_id, lid=model_id
        ),
        status=200,
        body=recommended_datetime_model_json,
    )
    recommendation = ModelRecommendation(project_id, model_id, RECOMMENDED_MODEL_TYPE.MOST_ACCURATE)
    model = recommendation.get_model()
    assert isinstance(model, DatetimeModel)
    assert hasattr(model, "model_type")
    assert hasattr(model, "effective_feature_derivation_window_start")
    assert model.effective_feature_derivation_window_start == -120
    assert model.effective_feature_derivation_window_end == 0
